import io
import mock
import os
import unittest
import zipfile

from cumulusci.core.config import OrgConfig
from cumulusci.tasks.salesforce.sourcetracking import ListChanges
from cumulusci.tasks.salesforce.sourcetracking import RetrieveChanges
from cumulusci.tasks.salesforce.sourcetracking import SnapshotChanges
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.utils import temporary_dir


class TestListChanges(unittest.TestCase):
    """List the changes from a scratch org"""

    def test_run_task(self):
        task = create_task(ListChanges, {"exclude": "Ignore"})
        task._init_task()
        task.tooling = mock.Mock()
        task.logger = mock.Mock()
        task.tooling.query_all.return_value = {
            "totalSize": 1,
            "records": [
                {
                    "MemberType": "CustomObject",
                    "MemberName": "Test__c",
                    "RevisionNum": 1,
                },
                {
                    "MemberType": "CustomObject",
                    "MemberName": "Ignored__c",
                    "RevisionNum": 2,
                },
            ],
        }
        task._run_task()
        self.assertIn("CustomObject: Test__c", task.logger.info.call_args[0][0])

    def test_run_task__no_changes(self):
        task = create_task(ListChanges)
        task._init_task()
        task.tooling = mock.Mock()
        task.logger = mock.Mock()
        task.tooling.query_all.return_value = {"totalSize": 0, "records": []}
        task._run_task()
        self.assertIn("Found no changes.", task.logger.info.call_args[0][0])

    def test_run_task__snapshot(self):
        with temporary_dir():
            task = create_task(ListChanges, {"snapshot": True})
            task._init_task()
            task.tooling = mock.Mock()
            messages = []
            task.logger = mock.Mock()
            task.logger.info = messages.append
            task.tooling.query_all.return_value = {
                "totalSize": 1,
                "records": [
                    {
                        "MemberType": "CustomObject",
                        "MemberName": "Test__c",
                        "RevisionNum": 1,
                    }
                ],
            }
            task._run_task()
            self.assertTrue(
                os.path.exists(
                    os.path.join(
                        ".sfdx", "orgs", "test-cci@example.com", "maxrevision.json"
                    )
                )
            )
            self.assertIn("CustomObject: Test__c", messages)

            task = create_task(ListChanges)
            task._init_task()
            task.tooling = mock.Mock()
            task.logger = mock.Mock()
            task.logger.info = messages.append
            task.tooling.query_all.return_value = {"totalSize": 0, "records": []}
            task._run_task()
            self.assertIn("Found no changes.", messages)

    def test_filter_changes__include(self):
        foo = {"MemberType": "CustomObject", "MemberName": "foo__c", "RevisionNum": 1}
        bar = {"MemberType": "CustomObject", "MemberName": "bar__c", "RevisionNum": 1}
        foobar = {
            "MemberType": "CustomObject",
            "MemberName": "foobar__c",
            "RevisionNum": 1,
        }
        task = create_task(ListChanges, {"include": "foo", "exclude": "bar"})
        filtered = task._filter_changes({"records": [foo, bar, foobar]})
        self.assertEqual([foo], filtered)


class TestRetrieveChanges(unittest.TestCase):
    """Retrieve changed components from a scratch org"""

    def test_run_task(self):
        with temporary_dir() as path:
            task = create_task(RetrieveChanges, {"path": path, "include": "Test"})
            task._init_task()
            task.tooling = mock.Mock()
            task.tooling.query_all.return_value = {
                "totalSize": 1,
                "records": [
                    {
                        "MemberType": "CustomObject",
                        "MemberName": "Test__c",
                        "RevisionNum": 1,
                    }
                ],
            }
            zf = zipfile.ZipFile(io.BytesIO(), "w")
            zf.writestr("objects/Test__c.object", "<root />")
            task.api_class = mock.Mock(return_value=mock.Mock(return_value=zf))
            task._run_task()
            with open(os.path.join(path, "package.xml"), "r") as f:
                package_xml = f.read()
        self.assertIn("<members>Test__c</members>", package_xml)

    def test_run_task__merge_changes(self):
        # If there is already an existing package,
        # we should add to it rather than overwriting it.
        with temporary_dir() as path:
            with open("package.xml", "w") as f:
                f.write(
                    """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <members>Object1</members>
        <name>CustomObject</name>
    </types>
</Package>
"""
                )
            task = create_task(RetrieveChanges, {"path": path})
            task._init_task()
            task.tooling = mock.Mock()
            task.tooling.query_all.return_value = {
                "totalSize": 1,
                "records": [
                    {
                        "MemberType": "CustomObject",
                        "MemberName": "Object2",
                        "RevisionNum": 1,
                    }
                ],
            }
            task.api_class = mock.Mock()
            task._run_task()

            package_xml = task.api_class.call_args[0][1]
            self.maxDiff = None
            self.assertEqual(
                """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <members>Object1</members>
        <members>Object2</members>
        <name>CustomObject</name>
    </types>
    <version>45.0</version>
</Package>""",
                package_xml,
            )

    def test_run_task__no_changes(self):
        with temporary_dir() as path:
            task = create_task(RetrieveChanges, {"path": path})
            task._init_task()
            messages = []
            task.tooling = mock.Mock()
            task.tooling.query_all.return_value = {"totalSize": 0, "records": []}
            task.logger = mock.Mock()
            task.logger.info = messages.append
            task._run_task()
            self.assertIn("No changes to retrieve", messages)


class TestSnapshotChanges(unittest.TestCase):
    def test_run_task(self):
        with temporary_dir():
            org_config = OrgConfig(
                {
                    "username": "test-cci@example.com",
                    "scratch": True,
                    "instance_url": "https://test.salesforce.com",
                    "access_token": "TOKEN",
                },
                "test",
            )
            task = create_task(SnapshotChanges, org_config=org_config)
            task._init_task()
            task.tooling.query = mock.Mock(return_value={"records": [{"num": 1}]})
            task._run_task()
            self.assertTrue(
                os.path.exists(
                    os.path.join(
                        ".sfdx", "orgs", "test-cci@example.com", "maxrevision.json"
                    )
                )
            )

    def test_freeze(self):
        task = create_task(SnapshotChanges)
        steps = task.freeze(None)
        self.assertEqual([], steps)



import os
from distutils.dir_util import copy_tree
from distutils.dir_util import remove_tree
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.tasks import BaseTask
from cumulusci.utils import findReplace


class CreateManagedSrc(BaseTask):
    task_options = {
        "path": {
            "description": "The path containing metadata to process for managed deployment",
            "required": True,
        },
        "revert_path": {
            "description": "The path to copy the original metadata to for the revert call",
            "required": True,
        },
    }

    managed_token = "//cumulusci-managed"

    def _run_task(self):
        # Check that path exists
        if not os.path.isdir(self.options["path"]):
            raise TaskOptionsError(
                "The path {} does not exist or is not a directory".format(
                    self.options["path"]
                )
            )

        # Check that revert_path does not exist
        if os.path.exists(self.options["revert_path"]):
            raise TaskOptionsError(
                "The revert_path {} already exists.  Delete it and try again".format(
                    self.options["revert_path"]
                )
            )

        # Copy path to revert_path
        copy_tree(self.options["path"], self.options["revert_path"])

        # Edit metadata in path
        self.logger.info(
            "Removing the string {0} from {1}/classes and {1}/triggers".format(
                self.managed_token, self.options["path"]
            )
        )
        findReplace(
            self.managed_token,
            "",
            os.path.join(self.options["path"], "classes"),
            "*.cls",
            self.logger,
        )
        findReplace(
            self.managed_token,
            "",
            os.path.join(self.options["path"], "triggers"),
            "*.trigger",
            self.logger,
        )

        self.logger.info(
            "{} has been stripped from all classes and triggers in {}".format(
                self.managed_token, self.options["path"]
            )
        )


class RevertManagedSrc(BaseTask):
    task_options = {
        "path": {
            "description": "The path containing metadata to process for managed deployment",
            "required": True,
        },
        "revert_path": {
            "description": "The path to copy the original metadata to for the revert call",
            "required": True,
        },
    }

    def _run_task(self):
        # Check that revert_path does exists
        if not os.path.isdir(self.options["revert_path"]):
            raise TaskOptionsError(
                "The revert_path {} does not exist or is not a directory".format(
                    self.options["revert_path"]
                )
            )

        self.logger.info(
            "Reverting {} from {}".format(
                self.options["path"], self.options["revert_path"]
            )
        )
        copy_tree(self.options["revert_path"], self.options["path"], update=1)
        self.logger.info("{} is now reverted".format(self.options["path"]))

        # Delete the revert_path
        self.logger.info("Deleting {}".format(self.options["revert_path"]))
        remove_tree(self.options["revert_path"])



"""Wraps the github3 library to configure request retries."""

from cumulusci.core.exceptions import GithubException
from github3 import login
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

retries = Retry(status_forcelist=(502, 503, 504), backoff_factor=0.3)
adapter = HTTPAdapter(max_retries=retries)


def get_github_api(username, password):
    gh = login(username, password)
    gh.session.mount("http://", adapter)
    gh.session.mount("https://", adapter)
    return gh


def validate_service(options):
    username = options["username"]
    password = options["password"]
    gh = get_github_api(username, password)
    try:
        gh.rate_limit()
    except Exception as e:
        raise GithubException(
            "Could not confirm access to the GitHub API: {}".format(str(e))
        )



""" a task for waiting on a Batch Apex job to complete """

import datetime
from cumulusci.utils import parse_api_datetime
from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.core.exceptions import SalesforceException

COMPLETED_STATUSES = ["Completed"]


class BatchApexWait(BaseSalesforceApiTask):
    """ BatchApexWait polls an org until the latest batch job
        for an apex class completes or fails """

    name = "BatchApexWait"
    batch = object()

    task_options = {
        "class_name": {
            "description": "Name of the Apex class to wait for.",
            "required": True,
        },
        "poll_interval": {
            "description": "Seconds to wait before polling for batch job completion. "
            "Defaults to 10 seconds."
        },
    }

    def _run_task(self):
        self.poll_interval_s = int(self.options.get("poll_interval", 10))

        self._poll()  # will block until poll_complete

        self.logger.info("Job is complete.")

        if not self.success:
            self.logger.info("There were some batch failures.")
            raise SalesforceException(self.batch["ExtendedStatus"])

        self.logger.info(
            "%s took %d seconds to process %d batches.",
            self.batch["ApexClass"]["Name"],
            self.delta,
            self.batch["TotalJobItems"],
        )

        return self.success

    def _poll_action(self):
        # get batch status
        query_results = self.tooling.query(self._batch_query)

        self.batch = query_results["records"][0]
        self.logger.info(
            "%s: %d of %d (%d failures)",
            self.batch["ApexClass"]["Name"],
            self.batch["JobItemsProcessed"],
            self.batch["TotalJobItems"],
            self.batch["NumberOfErrors"],
        )

        self.poll_complete = not self._poll_again()

    def _poll_again(self):
        return self.batch["Status"] not in COMPLETED_STATUSES

    @property
    def success(self):
        """ returns True if all batches succeeded """
        return (self.batch["JobItemsProcessed"] is self.batch["TotalJobItems"]) and (
            self.batch["NumberOfErrors"] is 0
        )

    @property
    def delta(self):
        """ returns the time (in seconds) that the batch took, if complete """
        completed_date = parse_api_datetime(self.batch["CompletedDate"])
        created_date = parse_api_datetime(self.batch["CreatedDate"])
        td = completed_date - created_date
        return td.total_seconds()

    @property
    def _batch_query(self):
        return (
            "SELECT Id, ApexClass.Name, Status, ExtendedStatus, TotalJobItems, "
            "JobItemsProcessed, NumberOfErrors, CreatedDate, CompletedDate "
            "FROM AsyncApexJob "
            "WHERE JobType='BatchApex' "
            "AND ApexClass.Name='{}' "
            "ORDER BY CreatedDate DESC "
            "LIMIT 1".format(self.options["class_name"])
        )



from __future__ import unicode_literals


class CumulusCIException(Exception):
    """ Base class for all CumulusCI Exceptions """

    pass


class CumulusCIUsageError(CumulusCIException):
    """ An exception thrown due to improper usage which should be resolvable by proper usage """

    pass


class CumulusCIFailure(CumulusCIException):
    """ An exception representing a failure such as a Metadata deployment failure or a test failure.  CI systems can handle these to determine fail vs error status """

    pass


class NotInProject(CumulusCIUsageError):
    """ Raised when no project can be found in the current context """

    pass


class ProjectConfigNotFound(CumulusCIUsageError):
    """ Raised when a project is found in the current context but no configuration was found for the project """

    pass


class ProjectMinimumVersionError(CumulusCIException):
    pass


class KeychainNotFound(CumulusCIException):
    """ Raised when no keychain could be found """

    pass


class KeychainKeyNotFound(CumulusCIUsageError):
    """ Raised when the keychain key couldn't be found """

    pass


class OrgNotFound(CumulusCIUsageError):
    """ Raised when no org could be found by a given name in the project keychain """

    pass


class ServiceNotConfigured(CumulusCIUsageError):
    """ Raised when no service configuration could be found by a given name in the project keychain """

    pass


class ServiceNotValid(CumulusCIUsageError):
    """ Raised when no service configuration could be found by a given name in the project configuration """

    pass


class DependencyResolutionError(CumulusCIException):
    """ Raised when an issue is encountered while resolving a static dependency map """

    pass


class ConfigError(CumulusCIException):
    """ Raised when a configuration enounters an error """

    def __init__(self, message=None, config_name=None):
        super(ConfigError, self).__init__(message)
        self.message = message
        self.config_name = config_name

    def __str__(self):
        return "{} for config {}".format(self.message, self.config_name)


class ConfigMergeError(ConfigError):
    """ Raised when merging configuration fails. """

    pass


class AntTargetException(CumulusCIException):
    """ Raised when a generic Ant target error occurs """

    pass


class DeploymentException(CumulusCIFailure):
    """ Raised when a metadata api deployment error occurs """

    pass


class ApexTestException(CumulusCIFailure):
    """ Raised when a build fails because of an Apex test failure """

    pass


class SalesforceCredentialsException(CumulusCIException):
    """ Raise when Salesforce credentials are invalid """

    pass


class TaskRequiresSalesforceOrg(CumulusCIUsageError):
    """ Raise when a task that requires a Salesforce org_config is not initialized with an org_config """

    pass


class TaskOptionsError(CumulusCIUsageError):
    """ Raise when a task's options are invalid """

    pass


class TaskNotFoundError(CumulusCIUsageError):
    """ Raise when task is not found in project config """

    pass


class FlowInfiniteLoopError(CumulusCIUsageError):
    """ Raised when a flow configuration creates a infinite loop """

    pass


class FlowConfigError(CumulusCIException):
    """ Raised when a flow configuration encounters an error """

    pass


class FlowNotFoundError(CumulusCIUsageError):
    """ Raise when flow is not found in project config """

    pass


class FlowNotReadyError(CumulusCIException):
    """ Raise when flow is called before it has been prepared """

    pass


class MrbelvedereError(CumulusCIException):
    """ Raise for errors from mrbelvedere installer """

    pass


class ScratchOrgException(CumulusCIException):
    """ Raise for errors related to scratch orgs """

    pass


class GithubException(CumulusCIException):
    """ Raise for errors related to GitHub """

    pass


class GithubApiError(CumulusCIException):
    pass


class GithubApiNotFoundError(CumulusCIException):
    pass


class GithubApiNoResultsError(CumulusCIException):
    pass


class GithubApiUnauthorized(CumulusCIException):
    pass


class SalesforceException(CumulusCIException):
    """ Raise for errors related to Salesforce """

    pass


class SalesforceDXException(CumulusCIException):
    """ Raise for errors related to Salesforce DX """

    pass


class SOQLQueryException(CumulusCIFailure):
    """ Raise for errors related to Salesforce DX """

    pass


class CommandException(CumulusCIFailure):
    """ Raise for errors coming from spawned CLI subprocesses """

    pass


class BrowserTestFailure(CumulusCIFailure):
    """ Raise when browser tests fail """

    pass


class ApexCompilationException(CumulusCIFailure):
    """ Raise when apex compilation fails """

    def __str__(self):
        line, problem = self.args
        return "Apex compilation failed on line {}: {}".format(line, problem)


class ApexException(CumulusCIFailure):
    """ Raise when an Apex Exception is raised in an org """

    def __str__(self):
        message, stacktrace = self.args
        stacktrace = "\n  ".join(stacktrace.splitlines())
        return "Apex error: {}\n  Stacktrace:\n  {}".format(message, stacktrace)


class PushApiObjectNotFound(CumulusCIException):
    """ Raise when Salesforce Push API object is not found """

    pass


class RobotTestFailure(CumulusCIFailure):
    """ Raise when a robot test fails in a test suite """

    pass


class BulkDataException(CumulusCIException):
    """ Raise for errors from bulkdata tasks """

    pass



import mock
import unittest

from cumulusci.core.config import ServiceConfig
from cumulusci.core.config import TaskConfig
from cumulusci.tasks.github import PullRequests
from cumulusci.tests.util import create_project_config


@mock.patch("cumulusci.tasks.github.base.get_github_api", mock.Mock())
class TestPullRequests(unittest.TestCase):
    project_config = create_project_config()
    project_config.keychain.set_service(
        "github",
        ServiceConfig(
            {
                "username": "TestUser",
                "password": "TestPass",
                "email": "testuser@testdomain.com",
            }
        ),
    )
    task_config = TaskConfig()
    task = PullRequests(project_config, task_config)
    repo = mock.Mock()
    repo.pull_requests.return_value = [mock.Mock(number=1, title="Test PR")]
    task.get_repo = mock.Mock(return_value=repo)
    task.logger = mock.Mock()
    task()
    task.logger.info.assert_called_with("#1: Test PR")



""" Tasks are the basic unit of execution in CumulusCI.

Subclass BaseTask or a descendant to define custom task logic
"""
from __future__ import division
from __future__ import unicode_literals

from builtins import object
from past.utils import old_div
import contextlib
import logging
import time
import threading

from cumulusci.core.exceptions import TaskRequiresSalesforceOrg
from cumulusci.core.exceptions import TaskOptionsError

CURRENT_TASK = threading.local()
CURRENT_TASK.stack = []


@contextlib.contextmanager
def stacked_task(self):
    CURRENT_TASK.stack.append(self)
    try:
        yield
    finally:
        CURRENT_TASK.stack.pop()


class BaseTask(object):
    """ BaseTask provides the core execution logic for a Task

    Subclass BaseTask and provide a `_run_task()` method with your
    code.
    """

    task_docs = ""
    task_options = {}
    salesforce_task = False  # Does this task require a salesforce org?

    def __init__(
        self,
        project_config,
        task_config,
        org_config=None,
        flow=None,
        name=None,
        stepnum=None,
        **kwargs
    ):
        self.project_config = project_config
        self.task_config = task_config
        self.org_config = org_config
        self.poll_count = 0
        self.poll_interval_level = 0
        self.poll_interval_s = 1
        self.poll_complete = False

        # dict of return_values that can be used by task callers
        self.return_values = {}

        # simple result object for introspection, often a return_code
        self.result = None

        # the flow for this task execution
        self.flow = flow

        # the tasks name in the flow
        self.name = name

        # the tasks stepnumber in the flow
        self.stepnum = stepnum

        self._init_logger()
        self._init_options(kwargs)
        self._validate_options()

    def _init_logger(self):
        """ Initializes self.logger """
        if self.flow:
            self.logger = self.flow.logger.getChild(self.__class__.__name__)
        else:
            self.logger = logging.getLogger(__name__)

    def _init_options(self, kwargs):
        """ Initializes self.options """
        self.options = self.task_config.options
        if self.options is None:
            self.options = {}
        if kwargs:
            self.options.update(kwargs)

        # Handle dynamic lookup of project_config values via $project_config.attr
        for option, value in list(self.options.items()):
            try:
                if value.startswith("$project_config."):
                    attr = value.replace("$project_config.", "", 1)
                    self.options[option] = getattr(self.project_config, attr, None)
            except AttributeError:
                pass

    def _validate_options(self):
        missing_required = []
        for name, config in list(self.task_options.items()):
            if config.get("required") is True and name not in self.options:
                missing_required.append(name)

        if missing_required:
            raise TaskOptionsError(
                "{} requires the options ({}) "
                "and no values were provided".format(
                    self.__class__.__name__, ", ".join(missing_required)
                )
            )

    def _update_credentials(self):
        """ Override to do any logic  to refresh credentials """
        pass

    def _init_task(self):
        """ Override to implement dynamic logic for initializing the task. """
        pass

    def __call__(self):
        # If sentry is configured, initialize sentry for error capture
        self.project_config.init_sentry()

        if self.salesforce_task and not self.org_config:
            raise TaskRequiresSalesforceOrg(
                "This task requires a salesforce org. "
                "Use org default <name> to set a default org "
                "or pass the org name with the --org option"
            )
        self._update_credentials()
        self._init_task()

        with stacked_task(self):
            try:
                self._log_begin()
                self.result = self._run_task()
                return self.return_values
            except Exception as e:
                self._process_exception(e)
                raise

    def _process_exception(self, e):
        if self.project_config.use_sentry:
            self.logger.info("Logging error to sentry.io")

            tags = {"task class": self.__class__.__name__}
            if self.org_config:
                tags["org username"] = self.org_config.username
                tags["scratch org"] = self.org_config.scratch is True
            for key, value in list(self.options.items()):
                tags["option_" + key] = value
            self.project_config.sentry.tags_context(tags)

            resp = self.project_config.sentry.captureException()
            self.project_config.sentry_event = resp

    def _run_task(self):
        """ Subclasses should override to provide their implementation """
        raise NotImplementedError("Subclasses should provide their own implementation")

    def _log_begin(self):
        """ Log the beginning of the task execution """
        self.logger.info("Beginning task: %s", self.__class__.__name__)
        if self.salesforce_task and not self.flow:
            self.logger.info("%15s %s", "As user:", self.org_config.username)
            self.logger.info("%15s %s", "In org:", self.org_config.org_id)
        self.logger.info("")

    def _retry(self):
        while True:
            try:
                self._try()
                break
            except Exception as e:
                if not (self.options["retries"] and self._is_retry_valid(e)):
                    raise
                if self.options["retry_interval"]:
                    self.logger.warning(
                        "Sleeping for {} seconds before retry...".format(
                            self.options["retry_interval"]
                        )
                    )
                    time.sleep(self.options["retry_interval"])
                    if self.options["retry_interval_add"]:
                        self.options["retry_interval"] += self.options[
                            "retry_interval_add"
                        ]
                self.options["retries"] -= 1
                self.logger.warning(
                    "Retrying ({} attempts remaining)".format(self.options["retries"])
                )

    def _try(self):
        raise NotImplementedError("Subclasses should provide their own implementation")

    def _is_retry_valid(self, e):
        return True

    def _poll(self):
        """ poll for a result in a loop """
        while True:
            self.poll_count += 1
            self._poll_action()
            if self.poll_complete:
                break
            time.sleep(self.poll_interval_s)
            self._poll_update_interval()

    def _poll_action(self):
        """
        Poll something and process the response.
        Set `self.poll_complete = True` to break polling loop.
        """
        raise NotImplementedError("Subclasses should provide their own implementation")

    def _poll_update_interval(self):
        """ update the polling interval to be used next iteration """
        # Increase by 1 second every 3 polls
        if old_div(self.poll_count, 3) > self.poll_interval_level:
            self.poll_interval_level += 1
            self.poll_interval_s += 1
            self.logger.info(
                "Increased polling interval to %d seconds", self.poll_interval_s
            )

    def freeze(self, step):
        ui_step = {
            "name": self.task_config.name or self.name,
            "kind": "other",
            "is_required": True,
        }
        ui_step.update(self.task_config.config.get("ui_options", {}))
        ui_step.update(
            {
                "path": step.path,
                "step_num": str(step.step_num),
                "task_class": self.task_config.class_path,
                "task_config": {"options": self.options},
            }
        )
        return [ui_step]



from builtins import str
from cumulusci.core.exceptions import ApexTestException
from cumulusci.core.exceptions import SalesforceException
from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class PackageUpload(BaseSalesforceApiTask):
    name = "PackageUpload"
    api_version = "38.0"
    task_options = {
        "name": {"description": "The name of the package version.", "required": True},
        "production": {
            "description": "If True, uploads a production release.  Defaults to uploading a beta"
        },
        "description": {
            "description": "A description of the package and what this version contains."
        },
        "password": {
            "description": "An optional password for sharing the package privately with anyone who has the password. Don't enter a password if you want to make the package available to anyone on AppExchange and share your package publicly."
        },
        "post_install_url": {
            "description": "The fully-qualified URL of the post-installation instructions. Instructions are shown as a link after installation and are available from the package detail view."
        },
        "release_notes_url": {
            "description": "The fully-qualified URL of the package release notes. Release notes are shown as a link during the installation process and are available from the package detail view after installation."
        },
        "namespace": {
            "description": "The namespace of the package.  Defaults to project__package__namespace"
        },
    }

    def _init_options(self, kwargs):
        super(PackageUpload, self)._init_options(kwargs)

        self.upload = None
        self.upload_id = None

        # Set the namespace option to the value from cumulusci.yml if not already set
        if not "namespace" in self.options:
            self.options["namespace"] = self.project_config.project__package__namespace

    def _run_task(self):
        package = self._get_one(
            "select Id from MetadataPackage where NamespacePrefix='{}'".format(
                self.options["namespace"]
            ),
            "No package found with namespace {}".format(self.options["namespace"]),
        )
        package_id = package["Id"]

        production = self.options.get("production", False) in [True, "True", "true"]
        package_info = {
            "VersionName": self.options["name"],
            "IsReleaseVersion": production,
            "MetadataPackageId": package_id,
        }

        if "description" in self.options:
            package_info["Description"] = self.options["description"]
        if "password" in self.options:
            package_info["Password"] = self.options["password"]
        if "post_install_url" in self.options:
            package_info["PostInstallUrl"] = self.options["post_install_url"]
        if "release_notes_url" in self.options:
            package_info["ReleaseNotesUrl"] = self.options["release_notes_url"]

        PackageUploadRequest = self._get_tooling_object("PackageUploadRequest")
        self.upload = PackageUploadRequest.create(package_info)
        self.upload_id = self.upload["id"]

        self.logger.info(
            "Created PackageUploadRequest {} for Package {}".format(
                self.upload_id, package_id
            )
        )
        self._poll()

        if self.upload["Status"] == "ERROR":
            self.logger.error("Package upload failed with the following errors")
            error = {"message": ""}
            for error in self.upload["Errors"]["errors"]:
                self.logger.error("  {}".format(error["message"]))

            # use the last error in the batch, but log them all.
            error_class = (
                ApexTestException
                if error["message"] == "ApexTestFailure"
                else SalesforceException
            )
            raise error_class("Package upload failed")
        else:
            version_id = self.upload["MetadataPackageVersionId"]
            version = self._get_one(
                "select MajorVersion, MinorVersion, PatchVersion, BuildNumber, ReleaseState from MetadataPackageVersion where Id = '{}'".format(
                    version_id
                ),
                "Version {} not found".format(version_id),
            )
            version_parts = [str(version["MajorVersion"]), str(version["MinorVersion"])]
            if version["PatchVersion"]:
                version_parts.append(str(version["PatchVersion"]))

            self.version_number = ".".join(version_parts)

            if version["ReleaseState"] == "Beta":
                self.version_number += " (Beta {})".format(version["BuildNumber"])

            self.return_values = {
                "version_number": str(self.version_number),
                "version_id": version_id,
                "package_id": package_id,
            }

            self.logger.info(
                "Uploaded package version {} with Id {}".format(
                    self.version_number, version_id
                )
            )

    def _get_one(self, query, message):
        result = self.tooling.query(query)
        if result["totalSize"] != 1:
            self.logger.error(message)
            raise SalesforceException(message)
        return result["records"][0]

    def _poll_action(self):
        soql_check_upload = "select Id, Status, Errors, MetadataPackageVersionId from PackageUploadRequest where Id = '{}'".format(
            self.upload_id
        )
        self.upload = self._get_one(
            soql_check_upload,
            "Failed to get info for upload with id {}".format(self.upload_id),
        )
        self.logger.info(
            "PackageUploadRequest {} is {}".format(
                self.upload_id, self.upload["Status"]
            )
        )

        self.poll_complete = not self._poll_again(self.upload["Status"])

    def _poll_again(self, upload_status):
        return upload_status in ["IN_PROGRESS", "QUEUED"]



import os
from distutils.dir_util import copy_tree
from distutils.dir_util import remove_tree
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.tasks import BaseTask
from cumulusci.utils import removeXmlElement


class CreateUnmanagedEESrc(BaseTask):
    task_options = {
        "path": {
            "description": "The path containing metadata to process for managed deployment",
            "required": True,
        },
        "revert_path": {
            "description": "The path to copy the original metadata to for the revert call",
            "required": True,
        },
    }

    elements = ["*.object:availableFields"]

    def _run_task(self):
        # Check that path exists
        if not os.path.isdir(self.options["path"]):
            raise TaskOptionsError(
                "The path {} does not exist or is not a directory".format(
                    self.options["path"]
                )
            )

        # Check that revert_path does not exist
        if os.path.exists(self.options["revert_path"]):
            raise TaskOptionsError(
                "The revert_path {} already exists.  Delete it and try again".format(
                    self.options["revert_path"]
                )
            )

        # Copy path to revert_path
        copy_tree(self.options["path"], self.options["revert_path"])

        # Edit metadata in path
        self.logger.info(
            "Preparing metadata in {0} for unmanaged EE deployment".format(
                self.options["path"]
            )
        )

        for element in self.elements:
            fname_match, element_name = element.split(":")
            removeXmlElement(
                element_name, os.path.join(self.options["path"]), fname_match
            )

        self.logger.info(
            "Metadata in {} is now prepared for unmanaged EE deployment".format(
                self.options["path"]
            )
        )


class RevertUnmanagedEESrc(BaseTask):
    task_options = {
        "path": {
            "description": "The path containing metadata to process for managed deployment",
            "required": True,
        },
        "revert_path": {
            "description": "The path to copy the original metadata to for the revert call",
            "required": True,
        },
    }

    def _run_task(self):
        # Check that revert_path does exists
        if not os.path.isdir(self.options["revert_path"]):
            raise TaskOptionsError(
                "The revert_path {} does not exist or is not a directory".format(
                    self.options["revert_path"]
                )
            )

        self.logger.info(
            "Reverting {} from {}".format(
                self.options["path"], self.options["revert_path"]
            )
        )
        copy_tree(self.options["revert_path"], self.options["path"], update=1)
        self.logger.info("{} is now reverted".format(self.options["path"]))

        # Delete the revert_path
        self.logger.info("Deleting {}".format(self.options["revert_path"]))
        remove_tree(self.options["revert_path"])



import unittest
from robot.libraries.BuiltIn import RobotNotRunningError
from cumulusci.robotframework.Salesforce import Salesforce
import mock


class TestLocators(unittest.TestCase):
    @mock.patch("cumulusci.robotframework.Salesforce.Salesforce.cumulusci")
    def test_locators_in_robot_context(self, cumulusci_mock):
        """Verify we can get locators for the current org api version"""
        cumulusci_mock.tooling._call_salesforce.return_value.json.return_value = [
            {"version": "45.0"}
        ]

        # This instantiates the robot library, mimicing a robot library import.
        # We've mocked out the code that would otherwise throw an error since
        # we're not running in the context of a robot test. The library should
        # return the latest version of the locators.
        sf = Salesforce()

        expected = "cumulusci.robotframework.locators_45"
        actual = sf.locators_module.__name__
        message = "expected to load '{}', actually loaded '{}'".format(expected, actual)
        self.assertEqual(expected, actual, message)

        pass

    @mock.patch(
        "robot.libraries.BuiltIn.BuiltIn.get_library_instance",
        side_effect=RobotNotRunningError(),
    )
    def test_locators_outside_robot_context(self, builtin_mock):
        """Verify that we get the latest locators if not running in the context of a robot test"""

        # This instantiates the robot library, mimicing a robot library import
        # however, because we've mocked get_library_instance to throw an error,
        # we expect the library to still be instantiated, but with the latest
        # version of the locators.
        sf = Salesforce()
        expected = "cumulusci.robotframework.locators_46"
        actual = sf.locators_module.__name__
        message = "expected to load '{}', actually loaded '{}'".format(expected, actual)
        self.assertEqual(expected, actual, message)

    def test_locators_46(self):
        """Verify that locators_46 is a superset of the locators_45

        This test is far from perfect, but it should at least flag a
        catastrophic error in how locators for a version augment the
        locators for previous versions.

        Note: this test assumes that locators_46 doesn't delete any of the
        keys from 45.

        """
        import cumulusci.robotframework.locators_45 as locators_45
        import cumulusci.robotframework.locators_46 as locators_46

        keys_45 = set(locators_45.lex_locators)
        keys_46 = set(locators_46.lex_locators)

        self.assertNotEqual(
            id(locators_45.lex_locators),
            id(locators_46.lex_locators),
            "locators_45.lex_locators and locators_46.lex_locators are the same object",
        )
        self.assertTrue(len(keys_45) > 0)
        self.assertTrue(keys_45.issubset(keys_46))



from future import standard_library

standard_library.install_aliases()
import http.client
import responses

from cumulusci.tasks.github.tests.util_github_api import GithubApiTestMixin


class MockUtil(GithubApiTestMixin):
    BASE_HTML_URL = "https://github.com"
    BASE_API_URL = "https://api.github.com"

    def __init__(self, owner, repo):
        self.owner = owner
        self.repo = repo
        self.init_github()

    @property
    def html_url(self):
        return "{}/{}/{}".format(self.BASE_HTML_URL, self.owner, self.repo)

    @property
    def repo_url(self):
        return "{}/repos/{}/{}".format(self.BASE_API_URL, self.owner, self.repo)

    def mock_edit_release(self, body=None, draft=True, prerelease=False):
        if body == None:
            body = "Test release body"
        responses.add(
            method=responses.PATCH,
            url="{}/releases/1".format(self.repo_url),
            json=self._get_expected_release(
                "1", body=body, draft=draft, prerelease=prerelease
            ),
            status=http.client.OK,
        )

    def mock_get_repo(self):
        responses.add(
            method=responses.GET,
            url=self.repo_url,
            json=self._get_expected_repo(self.owner, self.repo),
            status=http.client.OK,
        )

    def mock_list_pulls(self):
        responses.add(
            method=responses.GET,
            url="{}/pulls".format(self.repo_url),
            json=[{"id": 1, "number": 1}],
            status=http.client.OK,
        )

    def mock_get_release(self, tag, body):
        responses.add(
            method=responses.GET,
            url="{}/releases/tags/{}".format(self.repo_url, tag),
            json=self._get_expected_release(
                tag, url="{}/releases/1".format(self.repo_url), body=body
            ),
            status=http.client.OK,
        )

    def mock_post_comment(self, issue_number):
        responses.add(
            method=responses.POST,
            url="{}/issues/{}/comments".format(self.repo_url, issue_number),
            status=http.client.OK,
        )

    def mock_pull_request(self, pr_number, body, title=None):
        if title == None:
            title = "Test Pull Request Title"
        responses.add(
            method=responses.GET,
            url="{}/pulls/{}".format(self.repo_url, pr_number),
            json=self._get_expected_pull_request(pr_number, pr_number, body=body),
            status=http.client.OK,
        )



import sys

from cumulusci.core.config import BaseGlobalConfig, BaseProjectConfig
from cumulusci.core.exceptions import NotInProject, ProjectConfigNotFound
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.core.flowrunner import FlowCallback, FlowCoordinator


# pylint: disable=assignment-from-none
class BaseCumulusCI(object):
    global_config_class = BaseGlobalConfig
    project_config_class = BaseProjectConfig
    keychain_class = BaseProjectKeychain
    callback_class = FlowCallback

    def __init__(self, *args, **kwargs):
        load_project_config = kwargs.pop(
            "load_project_config", True
        )  # this & below can be added to fn signature in py3
        load_keychain = kwargs.pop("load_keychain", True)
        allow_global_keychain = kwargs.pop("allow_global_keychain", False)

        self.global_config = None
        self.project_config = None
        self.keychain = None
        self.is_global_keychain = False

        self._load_global_config()

        if load_project_config:
            try:
                self._load_project_config(*args, **kwargs)
                self._add_repo_to_path()
            except (NotInProject, ProjectConfigNotFound):
                if allow_global_keychain:
                    self.is_global_keychain = True
                else:
                    raise
            if load_keychain:
                self._load_keychain()

    @property
    def global_config_cls(self):
        klass = self.get_global_config_class()
        return klass or self.global_config_class

    def get_global_config_class(self):
        return None

    @property
    def project_config_cls(self):
        klass = self.get_project_config_class()
        return klass or self.project_config_class

    def get_project_config_class(self):
        return None

    @property
    def keychain_cls(self):
        klass = self.get_keychain_class()
        return klass or self.keychain_class

    def get_keychain_class(self):
        return None

    @property
    def keychain_key(self):
        return self.get_keychain_key()

    def get_keychain_key(self):
        return None

    def _add_repo_to_path(self):
        if self.project_config:
            sys.path.append(self.project_config.repo_root)

    def _load_global_config(self):
        self.global_config = self.global_config_cls()

    def _load_project_config(self, *args, **kwargs):
        self.project_config = self.project_config_cls(
            self.global_config, *args, **kwargs
        )

    def _load_keychain(self):
        if self.is_global_keychain:
            self.keychain = self.keychain_cls(self.global_config, self.keychain_key)
        else:
            self.keychain = self.keychain_cls(self.project_config, self.keychain_key)
            self.project_config.keychain = (
                self.keychain
            )  # never understood this but ok.

    def get_flow(self, name, options=None):
        """ Get a primed and readytogo flow coordinator. """
        config = self.project_config.get_flow(name)
        callbacks = self.callback_class()
        coordinator = FlowCoordinator(
            self.project_config,
            config,
            name=name,
            options=options,
            skip=None,
            callbacks=callbacks,
        )
        return coordinator



""" Tests for the SFDX Command Wrapper"""

import mock
import unittest

from mock import MagicMock
from mock import patch

from cumulusci.core.config import BaseGlobalConfig
from cumulusci.core.config import BaseProjectConfig
from cumulusci.core.config import TaskConfig
from cumulusci.core.config import OrgConfig
from cumulusci.core.config import ScratchOrgConfig
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.core.tests.utils import MockLoggerMixin

from cumulusci.tasks.command import CommandException
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.tasks.sfdx import SFDXBaseTask
from cumulusci.tasks.sfdx import SFDXOrgTask
from cumulusci.tasks.sfdx import SFDXJsonTask


class TestSFDXBaseTask(MockLoggerMixin, unittest.TestCase):
    """ Tests for the Base Task type """

    def setUp(self):
        self.global_config = BaseGlobalConfig()
        self.project_config = BaseProjectConfig(
            self.global_config, config={"noyaml": True}
        )
        self.task_config = TaskConfig()

        keychain = BaseProjectKeychain(self.project_config, "")
        self.project_config.set_keychain(keychain)

        self._task_log_handler.reset()
        self.task_log = self._task_log_handler.messages

    def test_base_task(self):
        """ The command is prefixed w/ sfdx """

        self.task_config.config["options"] = {"command": "force:org", "extra": "--help"}
        task = SFDXBaseTask(self.project_config, self.task_config)

        try:
            task()
        except CommandException:
            pass

        self.assertEqual("sfdx force:org --help", task.options["command"])

    @patch(
        "cumulusci.tasks.sfdx.SFDXOrgTask._update_credentials",
        MagicMock(return_value=None),
    )
    @patch("cumulusci.tasks.command.Command._run_task", MagicMock(return_value=None))
    def test_keychain_org_creds(self):
        """ Keychain org creds are passed by env var """

        self.task_config.config["options"] = {"command": "force:org --help"}
        access_token = "00d123"
        org_config = OrgConfig(
            {
                "access_token": access_token,
                "instance_url": "https://test.salesforce.com",
            },
            "test",
        )

        task = SFDXOrgTask(self.project_config, self.task_config, org_config)

        try:
            task()
        except CommandException:
            pass

        self.assertIn("SFDX_INSTANCE_URL", task._get_env())
        self.assertIn("SFDX_USERNAME", task._get_env())
        self.assertIn(access_token, task._get_env()["SFDX_USERNAME"])

    def test_scratch_org_username(self):
        """ Scratch Org credentials are passed by -u flag """
        self.task_config.config["options"] = {"command": "force:org --help"}
        org_config = ScratchOrgConfig({"username": "test@example.com"}, "test")

        task = SFDXOrgTask(self.project_config, self.task_config, org_config)
        self.assertIn("-u test@example.com", task.options["command"])


class TestSFDXJsonTask(unittest.TestCase):
    def test_get_command(self):
        task = create_task(SFDXJsonTask)
        command = task._get_command()
        self.assertEqual("sfdx force:mdapi:deploy --json", command)

    def test_process_output(self):
        task = create_task(SFDXJsonTask)
        task.logger = mock.Mock()
        task._process_output("{}")
        task.logger.info.assert_called_once_with("JSON = {}")

    def test_process_output__json_parse_error(self):
        task = create_task(SFDXJsonTask)
        task.logger = mock.Mock()
        task._process_output("{")
        task.logger.error.assert_called_once()



import logging
import sarge

logger = logging.getLogger(__name__)


def sfdx(command, username=None, log_note=None):
    """Call an sfdx command and capture its output.

    Be sure to quote user input that is part of the command using `sarge.shell_format`.

    Returns a `sarge` Command instance with returncode, stdout, stderr
    """
    command = "sfdx {}".format(command)
    if username:
        command += sarge.shell_format(" -u {0}", username)
    if log_note:
        logger.info("{} with command: {}".format(log_note, command))
    p = sarge.Command(
        command,
        stdout=sarge.Capture(buffer_size=-1),
        stderr=sarge.Capture(buffer_size=-1),
        shell=True,
    )
    p.run()
    return p



import github3.exceptions

from cumulusci.core.utils import import_global
from cumulusci.tasks.release_notes.exceptions import CumulusCIException
from cumulusci.tasks.release_notes.parser import ChangeNotesLinesParser
from cumulusci.tasks.release_notes.parser import IssuesParser
from cumulusci.tasks.release_notes.provider import StaticChangeNotesProvider
from cumulusci.tasks.release_notes.provider import DirectoryChangeNotesProvider
from cumulusci.tasks.release_notes.provider import GithubChangeNotesProvider


class BaseReleaseNotesGenerator(object):
    def __init__(self):
        self.change_notes = []
        self.init_parsers()
        self.init_change_notes()

    def __call__(self):
        self._parse_change_notes()
        return self.render()

    def init_change_notes(self):
        self.change_notes = self._init_change_notes()

    def _init_change_notes(self):
        """ Subclasses should override this method to return an initialized
        subclass of BaseChangeNotesProvider """
        return []

    def init_parsers(self):
        """ Initializes the parser instances as the list self.parsers """
        self.parsers = []
        self._init_parsers()

    def _init_parsers(self):
        """ Subclasses should override this method to initialize their
        parsers """
        pass

    def _parse_change_notes(self):
        """ Parses all change_notes in self.change_notes() through all parsers
        in self.parsers """
        for change_note in self.change_notes():
            self._parse_change_note(change_note)

    def _parse_change_note(self, change_note):
        """ Parses an individual change note through all parsers in
        self.parsers """
        for parser in self.parsers:
            parser.parse(change_note)

    def render(self):
        """ Returns the rendered release notes from all parsers as a string """
        release_notes = []
        for parser in self.parsers:
            parser_content = parser.render()
            if parser_content is not None:
                release_notes.append(parser_content)
        return u"\r\n\r\n".join(release_notes)


class StaticReleaseNotesGenerator(BaseReleaseNotesGenerator):
    def __init__(self, change_notes):
        self._change_notes = change_notes
        super(StaticReleaseNotesGenerator, self).__init__()

    def _init_parsers(self):
        self.parsers.append(ChangeNotesLinesParser(self, "Critical Changes"))
        self.parsers.append(ChangeNotesLinesParser(self, "Changes"))
        self.parsers.append(IssuesParser(self, "Issues Closed"))

    def _init_change_notes(self):
        return StaticChangeNotesProvider(self, self._change_notes)


class DirectoryReleaseNotesGenerator(BaseReleaseNotesGenerator):
    def __init__(self, directory):
        self.directory = directory
        super(DirectoryReleaseNotesGenerator, self).__init__()

    def _init_parsers(self):
        self.parsers.append(ChangeNotesLinesParser(self, "Critical Changes"))
        self.parsers.append(ChangeNotesLinesParser(self, "Changes"))
        self.parsers.append(IssuesParser(self, "Issues Closed"))

    def _init_change_notes(self):
        return DirectoryChangeNotesProvider(self, self.directory)


class GithubReleaseNotesGenerator(BaseReleaseNotesGenerator):
    def __init__(
        self,
        github,
        github_info,
        parser_config,
        current_tag,
        last_tag=None,
        link_pr=False,
        publish=False,
        has_issues=True,
    ):
        self.github = github
        self.github_info = github_info
        self.parser_config = parser_config
        self.current_tag = current_tag
        self.last_tag = last_tag
        self.link_pr = link_pr
        self.do_publish = publish
        self.has_issues = has_issues
        self.lines_parser_class = None
        self.issues_parser_class = None
        super(GithubReleaseNotesGenerator, self).__init__()

    def __call__(self):
        release = self._get_release()
        content = super(GithubReleaseNotesGenerator, self).__call__()
        content = self._update_release_content(release, content)
        if self.do_publish:
            release.edit(body=content)
        return content

    def _init_parsers(self):
        for cfg in self.parser_config:
            parser_class = import_global(cfg["class_path"])
            self.parsers.append(parser_class(self, cfg["title"]))

    def _init_change_notes(self):
        return GithubChangeNotesProvider(self, self.current_tag, self.last_tag)

    def _get_release(self):
        repo = self.get_repo()
        try:
            return repo.release_from_tag(self.current_tag)
        except github3.exceptions.NotFoundError:
            raise CumulusCIException(
                "Release not found for tag: {}".format(self.current_tag)
            )

    def _update_release_content(self, release, content):
        """Merge existing and new release content."""
        if release.body:
            new_body = []
            current_parser = None
            is_start_line = False
            for parser in self.parsers:
                parser.replaced = False

            # update existing sections
            for line in release.body.splitlines():

                if current_parser:
                    if current_parser._is_end_line(current_parser._process_line(line)):
                        parser_content = current_parser.render()
                        if parser_content:
                            # replace existing section with new content
                            new_body.append(parser_content + "\r\n")
                        current_parser = None

                for parser in self.parsers:
                    if (
                        parser._render_header().strip()
                        == parser._process_line(line).strip()
                    ):
                        parser.replaced = True
                        current_parser = parser
                        is_start_line = True
                        break
                    else:
                        is_start_line = False

                if is_start_line:
                    continue
                if current_parser:
                    continue
                else:
                    # preserve existing sections
                    new_body.append(line.strip())

            # catch section without end line
            if current_parser:
                new_body.append(current_parser.render())

            # add new sections at bottom
            for parser in self.parsers:
                parser_content = parser.render()
                if parser_content and not parser.replaced:
                    new_body.append(parser_content + "\r\n")

            content = u"\r\n".join(new_body)

        return content

    def get_repo(self):
        return self.github.repository(
            self.github_info["github_owner"], self.github_info["github_repo"]
        )
